from __future__ import annotations

from unittest.mock import Mock

import pytest

from pyplz import plz
from pyplz.command import Command
from tests.conftest import TestUtils


class TestPlzDefault:
    @TestUtils.patch_method(plz.list_tasks)
    def test_default_behavior(self, mock_list_tasks):
        plz._main_execute(Command())
        mock_list_tasks.assert_called_once()

    def test_default_custom_task(self):
        mock_func = Mock()

        @plz.task(default=True)
        def sample_task():
            mock_func()

        cmd = Command()
        plz._main_execute(cmd)

        mock_func.assert_called_once()

    def test_multiple_default_tasks_error(self):
        mock_func = Mock()

        @plz.task(default=True)
        def sample_task_1():
            mock_func()

        @plz.task(default=True)
        def sample_task_2():
            mock_func()

        with pytest.raises(SystemExit):
            # multiple defaults should raise an error
            cmd = Command()
            plz._main_execute(cmd)



from __future__ import annotations

from unittest.mock import Mock, call


from pyplz import plz
from pyplz.command import Command


class TestPlzTaskRequires:
    def test_task_requires_single(self):
        mock_parent = Mock()
        mock_parent.required, mock_parent.requires = Mock(), Mock()

        @plz.task()
        def required():
            mock_parent.required()

        @plz.task(requires=required)
        def requires():
            mock_parent.requires()

        cmd = Command("requires")
        plz._main_execute(cmd)

        # Assert the order of calls
        mock_parent.assert_has_calls([call.required(), call.requires()])

    def test_task_requires_single_with_args(self):
        mock_parent = Mock()
        mock_parent.required, mock_parent.requires = Mock(), Mock()

        @plz.task()
        def required(arg1, arg2):
            mock_parent.required(arg1, arg2)

        @plz.task(requires=[(required, ("arg1_value", "arg2_value"))])
        def requires():
            mock_parent.requires()

        cmd = Command("requires")
        plz._main_execute(cmd)

        # Assert the order of calls
        mock_parent.assert_has_calls([call.required("arg1_value", "arg2_value"), call.requires()])

    def test_task_requires_multiple_with_args(self):
        mock_parent = Mock()
        mock_parent.required1, mock_parent.required1, mock_parent.requires = Mock(), Mock(), Mock()

        @plz.task()
        def required1(arg1, arg2):
            mock_parent.required1(arg1, arg2)

        @plz.task()
        def required2(arg1, arg2):
            mock_parent.required2(arg1, arg2)

        @plz.task(requires=[(required1, ("req1_arg1_v", "req1_arg2_v")), (required2, ("req2_arg1_v", "req2_arg2_v"))])
        def requires():
            mock_parent.requires()

        cmd = Command("requires")
        plz._main_execute(cmd)

        # Assert the order of calls
        mock_parent.assert_has_calls(
            [
                call.required1("req1_arg1_v", "req1_arg2_v"),
                call.required2("req2_arg1_v", "req2_arg2_v"),
                call.requires(),
            ]
        )

    def test_task_requires_multiple(self):
        mock_parent = Mock()
        mock_parent.required1, mock_parent.required1, mock_parent.requires = Mock(), Mock(), Mock()

        @plz.task()
        def required1():
            mock_parent.required1()

        @plz.task()
        def required2():
            mock_parent.required2()

        @plz.task(requires=[required1, required2])
        def requires():
            mock_parent.requires()

        cmd = Command("requires")
        plz._main_execute(cmd)

        # Assert the order of calls
        mock_parent.assert_has_calls(
            [
                call.required1(),
                call.required2(),
                call.requires(),
            ]
        )

    def test_task_requires_chain_dep(self):
        mock_parent = Mock()
        mock_parent.required1, mock_parent.required1, mock_parent.requires = Mock(), Mock(), Mock()

        @plz.task()
        def required1():
            mock_parent.required1()

        @plz.task(requires=required1)
        def required2():
            mock_parent.required2()

        @plz.task(requires=[required1, required2])
        def requires():
            mock_parent.requires()

        cmd = Command("requires")
        plz._main_execute(cmd)

        # Assert the order of calls
        mock_parent.assert_has_calls(
            [
                call.required1(),
                call.required1(),
                call.required2(),
                call.requires(),
            ]
        )



from __future__ import annotations

from pyplz import plz
from pyplz.command import Command


class TestPlzList:
    def test_list_without_tasks(self, capfd):
        cmd = Command(list=True)

        plz._main_execute(cmd)

        assert len(plz._tasks) == 0
        out = capfd.readouterr().out
        assert "No tasks have been registered" in out



from __future__ import annotations

from unittest.mock import Mock

from pyplz import plz
from pyplz.command import Command


class TestPlzTask:
    # region wrapped functions for pure functionality

    def test_wrapped_func_without_arguments(self):
        @plz.task()
        def sample_task():
            return "Task executed"

        result = sample_task()
        assert result == "Task executed"

    def test_wrapped_func_with_arguments(self):
        @plz.task()
        def sample_task(arg1, arg2):
            return f"Task executed with {arg1} and {arg2}"

        result = sample_task("arg1_value", "arg2_value")
        assert result == "Task executed with arg1_value and arg2_value"

    def test_wrapped_func_with_default_arguments(self):
        @plz.task()
        def sample_task(arg1="default1", arg2="default2"):
            return f"Task executed with {arg1} and {arg2}"

        result = sample_task()
        assert result == "Task executed with default1 and default2"

    # endregion

    def test_task_without_arguments(self):
        mock_func = Mock()

        @plz.task()
        def sample_task():
            mock_func()

        cmd = Command("sample_task")
        plz._main_execute(cmd)

        mock_func.assert_called_once()

    def test_task_with_arguments(self):
        mock_func = Mock()

        @plz.task()
        def sample_task(arg1, arg2):
            mock_func(arg1, arg2)

        cmd = Command("sample_task", _args=["arg1_value", "arg2_value"])
        plz._main_execute(cmd)

        mock_func.assert_called_once_with("arg1_value", "arg2_value")

    def test_task_with_positional_arguments(self):
        mock_func = Mock()

        @plz.task()
        def sample_task(*args):
            mock_func(*args)

        cmd = Command("sample_task", _args=["arg1_value", "arg2_value"])
        plz._main_execute(cmd)

        mock_func.assert_called_once_with("arg1_value", "arg2_value")

    def test_task_with_default_arguments(self):
        mock_func = Mock()

        @plz.task()
        def sample_task(arg1="default1", arg2="default2"):
            mock_func(arg1, arg2)

        cmd = Command("sample_task")
        plz._main_execute(cmd)

        mock_func.assert_called_once_with("default1", "default2")

    def test_task_prints_return_value(self, capfd):
        @plz.task()
        def sample_task():
            return "Task executed"

        cmd = Command("sample_task")
        plz._main_execute(cmd)

        captured = capfd.readouterr()
        assert "Task executed" in captured.out



import sys

from pyplz.error_handler import ErrorHandler
from pyplz.plz_app import plz
from pyplz.plz_parser import PlzParser
from pyplz.task_parser import TaskParser


def main():
    with ErrorHandler() as h:
        plz._load_plzfile()
        parser = PlzParser(plz_app=plz, task_parser=TaskParser())
        command = parser.parse(sys.argv[1:])

    if h.caught:
        sys.exit(1)

    plz._main_execute(command)


if __name__ == "__main__":
    main()



from unittest.mock import Mock

import pytest

from pyplz import plz
from pyplz.exceptions import TaskNotFoundError
from pyplz.plz_parser import PlzParser


class TestPlzParser:
    def test_plz_parser_task_with_args_task_parser_called(self):
        @plz.task()
        def sample_task(some_bool: bool):
            pass

        task_parser_mock = Mock()
        task_parser_mock.parse = Mock()
        parser = PlzParser(plz_app=plz, task_parser=task_parser_mock)

        parser.parse(["sample_task", "--some-bool"])

        task_parser_mock.parse.assert_called_once_with(task=plz._tasks["sample_task"], args=["--some-bool"])

    def test_plz_parser_task_without_args_task_parser_called(self):
        @plz.task()
        def sample_task(some_bool: bool):
            pass

        task_parser_mock = Mock()
        task_parser_mock.parse = Mock()
        parser = PlzParser(plz_app=plz, task_parser=task_parser_mock)

        parser.parse(["sample_task"])

        task_parser_mock.parse.assert_called_once_with(task=plz._tasks["sample_task"], args=[])

    def test_plz_parser_nonexisting_task_error_raised(self):
        task_parser_mock = Mock()
        task_parser_mock.parse = Mock()
        parser = PlzParser(plz_app=plz, task_parser=task_parser_mock)

        with pytest.raises(TaskNotFoundError) as einfo:
            parser.parse(["non_existing_task"])
            assert einfo.value.task_name == "non_existing_task"

    def test_plz_parser_empty(self):
        task_parser_mock = Mock()
        task_parser_mock.parse = Mock()
        parser = PlzParser(plz_app=plz, task_parser=task_parser_mock)

        cmd = parser.parse([])

        assert cmd.is_default()
        assert cmd.task is None
        assert cmd.task_kwargs is None
        assert cmd.list is False
        assert cmd.list_env is False
        assert cmd.list_env_all is False



from typing import Optional
from unittest.mock import Mock

import pytest
from pyplz import plz
from pyplz.exceptions import ForwardRefrenceNotSupported
from pyplz.task_parser import TaskParser


class TestTaskParser:
    def test_task_parser_kwargs_single(self):
        @plz.task()
        def sample_task(some_bool: bool):
            pass

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        cmd = parser.parse(task=task, args=["--some-bool"])

        assert cmd.task == task
        assert cmd.task_kwargs == {"some_bool": True}

    def test_task_parser_kwargs_default_true(self):
        @plz.task()
        def sample_task(some_bool: bool = True):
            pass

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        cmd = parser.parse(task=task, args=["--some-bool"])

        assert cmd.task == task
        assert cmd.task_kwargs == {"some_bool": False}

    def test_task_parser_kwargs_multi(self):
        impl_mock = Mock()

        @plz.task()
        def sample_task(some_bool: bool, some_int: int):
            impl_mock(some_bool=some_bool, some_int=some_int)

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        cmd = parser.parse(task=task, args=["--some-bool", "--some-int", "42"])

        assert cmd.task == task
        assert cmd.task_kwargs == {"some_bool": True, "some_int": 42}

    def test_task_parser_vanilla(self):
        impl_mock = Mock()

        @plz.task()
        def sample_task():
            impl_mock()

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        cmd = parser.parse(task=task, args=[])

        assert cmd.task == task
        assert cmd.task_kwargs == {}

    def test_task_parser_vanilla_missing_args(self, capfd):
        @plz.task()
        def sample_task(some_int: int):
            pass

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        with pytest.raises(SystemExit) as exc_info:
            parser.parse(task=task, args=[])

        assert exc_info.value.code == 2  # argparse exits with code 2 on error

        # Capture the output
        captured = capfd.readouterr()
        assert "the following arguments are required: --some-int" in captured.err

    def test_task_parser_vanilla_missing_args_but_has_default(self, capfd):
        @plz.task()
        def sample_task(some_int: int = 10):
            pass

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        cmd = parser.parse(task=task, args=[])

        assert cmd.task == task
        assert cmd.task_kwargs == {"some_int": 10}

    def test_task_parser_vanilla_missing_args_but_nullable(self, capfd):
        @plz.task()
        def sample_task(some_int: Optional[int]):
            pass

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        cmd = parser.parse(task=task, args=[])

        assert cmd.task == task
        assert cmd.task_kwargs == {"some_int": None}

    def test_task_parser_forward_ref_causes_error(self, capfd):
        @plz.task()
        def sample_task(some_int: "int"):
            pass

        task = plz._tasks["sample_task"]
        parser = TaskParser()

        with pytest.raises(ForwardRefrenceNotSupported):
            parser.parse(task=task, args=[])



from __future__ import annotations

import inspect
import os

from typing import Any, Callable

from rich.console import Console


console = Console()


class Task:
    def __init__(
        self,
        func: Callable,
        name: str | None = None,
        desc: str | None = None,
        requires: list[tuple[Task, tuple]] | None = None,  # List of tuples of tasks and their arguments
        is_default: bool = False,
        is_builtin: bool = False,
        task_env_vars: dict[str, Any] | None = None,
    ) -> None:
        self.func = func
        self.name = name or func.__name__
        desc = desc or ""
        if len(desc) == 0 and func.__doc__ is not None:
            desc = inspect.cleandoc(func.__doc__)
        self.desc = desc
        self.is_default = is_default
        self.is_builtin = is_builtin

        # normalize requires
        if requires is None:
            requires = []
        self.requires = requires

        # normalize task-scope environment variables
        if task_env_vars is None:
            task_env_vars = {}
        self.task_env_vars = task_env_vars

    def __call__(self, *args, **kwargs) -> Any:
        """
        Call the task function and return the result.
        If the task function has any required functions, they will be called first (recursively).
        #"""
        # load task-level environment variables
        for key, value in self.task_env_vars.items():
            os.environ[key] = value

        # Invoke required tasks first
        # for r_task, r_args in self.requires:
        #     r_task(*r_args)

        ret = self.func(*args, **kwargs)
        if ret is not None:
            console.print(ret)

    def __str__(self) -> str:
        tags = []
        if self.is_default:
            tags.append("[default]")
        if self.is_builtin:
            tags.append("[builtin]")
        tags_str = (" " + " ".join(tags)) if tags else ""
        return self.name + tags_str

    def __repr__(self) -> str:
        return self.__str__()



import sys
from unittest.mock import Mock

from pyplz import plz
from pyplz.main import main


class TestMain:
    def test_main_task(self):
        impl_mock = Mock()

        @plz.task()
        def sample_task():
            impl_mock()

        sys.argv = ["plz", "sample_task"]

        main()

        impl_mock.assert_called_once()

    def test_main_nonexisting_task(self):
        impl_mock = Mock()

        @plz.task()
        def sample_task():
            impl_mock()

        sys.argv = ["plz", "nonexisting_task"]

        main()

    def test_main_task_with_missing_argument(self):
        impl_mock = Mock()

        @plz.task()
        def sample_task(some_bool: bool, some_int: int):
            impl_mock()

        sys.argv = ["plz", "sample_task", "--some-bool"]

        main()
        # impl_mock.assert_called_once_with(some_bool=True)



from abc import ABC, abstractmethod


class PlzError(Exception, ABC):
    @abstractmethod
    def verbose(self):
        pass


class ParsingError(PlzError):
    pass


class TaskDefinitionError(PlzError):
    pass


class PlzFileError(PlzError):
    pass


class TaskNotFoundError(ParsingError):
    def __init__(self, task_name: str):
        self.task_name = task_name

    def verbose(self):
        return f"Task not found: `{self.task_name}`"


class TypeAnnotationNotAllowed(TaskDefinitionError):
    def __init__(self, param_type_hint: str):
        self.param_type_hint = param_type_hint

    def verbose(self):
        return f"Type annotation `{self.param_type_hint}` is not allowed."


class ForwardRefrenceNotSupported(PlzFileError):
    def verbose(self):
        return (
            "Forward reference is not supported in type plzfile.\n"
            "This can happen when using `from __future__ import annotations`\n"
            "Or using a string as a type hint.\n"
            "Please use the actual type instead."
        )



import argparse
import inspect
from typing import Union, get_args, get_origin

from pyplz.command import Command
from pyplz.exceptions import ForwardRefrenceNotSupported
from pyplz.task import Task


class TaskParser:
    def __init__(self):
        self._parser = argparse.ArgumentParser(add_help=False)

    @staticmethod
    def _is_nullable(param_type) -> bool:
        origin = get_origin(param_type)
        if origin is Union:
            return type(None) in get_args(param_type)
        return False

    def _add_task_arguments(self, task: Task):
        """Dynamically add arguments to the task parser based on the task's signature."""
        sig = inspect.signature(task.func)
        self._parser.description = task.desc
        g_system = self._parser.add_argument_group(title="system options")
        g_required = self._parser.add_argument_group(title="required arguments")
        g_optional = self._parser.add_argument_group(title="optional arguments")
        g_system.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )
        for name, param in sig.parameters.items():
            arg_name = f"--{name.replace('_', '-')}"
            param_type = param.annotation
            if type(param_type) is str:
                raise ForwardRefrenceNotSupported()
            is_nullable = self._is_nullable(param_type)
            has_default = param.default != inspect.Parameter.empty
            # we allow not setting an arg if it's optional or has a default value
            is_required = not is_nullable and not has_default

            kwargs = {}
            brackets_info = []
            brackets_info.append(param_type.__name__)

            # annotation is the type hint, can either be a type or a string
            # (in case of forward references using __future__)
            if param.annotation is bool:
                help = f"Set {name} to true"
                # bool is a special case, as it's always optional
                is_required = False
                if has_default and param.default is True:
                    brackets_info.append("default: true")
                    default = True
                    action = "store_false"
                else:
                    brackets_info.append("default: false")
                    default = False
                    action = "store_true"
                kwargs["default"] = default
            else:
                help = f"Set {name} to the provided value"
                action = "store"
                if has_default:
                    default = param.default
                    kwargs["default"] = default
                    brackets_info.append(f"default: {default}")
                kwargs["type"] = param_type

            help += f" ({', '.join(brackets_info)})"

            group = g_required if is_required else g_optional

            group.add_argument(
                arg_name,
                action=action,
                help=help,
                required=is_required,
                **kwargs,
            )

    def parse(self, task: Task, args: list[str]):
        self._add_task_arguments(task)
        parsed_args = self._parser.parse_args(args)
        kwargs = {k.replace("-", "_"): v for k, v in vars(parsed_args).items()}
        return Command(
            task=task,
            task_kwargs=kwargs,
        )



import pytest
from pyplz.task import Task


class TestTask:
    def test_task_desc_multiline(self):
        def test_task():
            """
            This is a test task
            that spans multiple lines.
            Let's see if it works.
            """
            pass

        task = Task(test_task)

        assert task.desc == "This is a test task\nthat spans multiple lines.\nLet's see if it works."

    def test_task_desc(self):
        def test_task():
            """This is a test task"""
            pass

        task = Task(test_task)

        assert task.desc == "This is a test task"

    def test_task_requires_normalized(self):
        def test_task():
            pass

        task = Task(test_task, requires=None)

        assert task.requires == []

    def test_task_called_with_missing_args(self):
        def test_task(arg1, arg2):
            pass

        task = Task(test_task)

        with pytest.raises(SystemExit):
            task("arg1")

    def test_task_to_string(self):
        def sample_task():
            pass

        task = Task(sample_task)

        assert str(task) == "sample_task"

    def test_task_to_string_default(self):
        def sample_task():
            pass

        task = Task(sample_task, is_default=True)

        assert str(task) == "sample_task [default]"

    def test_task_to_string_builtin(self):
        def sample_task():
            pass

        task = Task(sample_task, is_builtin=True)

        assert str(task) == "sample_task [builtin]"

    def test_task_to_string_default_builtin(self):
        def sample_task():
            pass

        task = Task(sample_task, is_default=True, is_builtin=True)

        assert str(task) == "sample_task [default] [builtin]"

    def test_task_print_help(self, capfd):
        def sample_task():
            pass

        task = Task(sample_task, desc="This is a sample task description.")
        task.print_help()

        # Capture the output
        captured = capfd.readouterr()

        # Check that the function name is present
        assert "sample_task" in captured.out

        # Check that the docstring is present
        assert "This is a sample task" in captured.out

        # Check that the arguments are present
        assert "param1" in captured.out
        assert "param2" in captured.out
        assert "default" in captured.out



from pyplz.exceptions import PlzError
from pyplz.plz_app import PlzApp


class ErrorHandler:
    def __init__(self):
        self.caught = False

    def __enter__(self):
        # Any setup can be done here if needed
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            return False

        if issubclass(exc_type, PlzError):
            PlzApp.print_error(exc_value.verbose(), exit=True)

        # False to raise the unhandled exception
        return False



import sys
from unittest.mock import Mock

from pytest import skip
import pytest

from pyplz.command import Parser
from pyplz.main import main
from pyplz.plz_app import PlzApp
from tests.conftest import TestUtils


class TestMain:
    @TestUtils.patch_method(PlzApp._main_execute)
    def test_main_configured_called(self, mock_main_execute):
        sys.argv = ["pyplz", "test"]
        main()
        mock_main_execute.assert_called_once()

    @TestUtils.patch_method(PlzApp._main_execute)
    def test_main_main_execute_called(self, mock_main_execute):
        sys.argv = ["pyplz", "test"]
        main()
        mock_main_execute.assert_called_once()

    @pytest.mark.skip("Not implemented")
    @TestUtils.patch_method(PlzApp._main_execute)
    def test_main_parsed_command_executed(self, mock_parse_args, mock_main_execute):
        mocked_command = Mock()
        mock_parse_args.return_value = mocked_command
        sys.argv = ["pyplz", "test"]
        main()
        mock_parse_args.assert_called_once_with(["test"])
        mock_main_execute.assert_called_once_with(mocked_command)



from __future__ import annotations

import importlib.util
import inspect
import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Any, Callable

from dotenv import dotenv_values, load_dotenv
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from pyplz.command import Command
from pyplz.console_utils import ConsoleUtils
from pyplz.task import Task
from pyplz.types import CallableWithArgs

console = Console()


class PlzApp:
    def __init__(self) -> None:
        self._tasks: dict[str, Task] = dict()
        self._user_configured = False

    def _reset(self):
        self._tasks = dict()
        self._user_configured = False

    def configure(self, env: dict[str, str] | None = None):
        self._user_configured = True
        self._dotenv = self._load_env_dotenv()
        # plz loads the CWD to the sys.path to allow plzfile to import freely
        sys.path.append(os.path.join(os.getcwd()))

        if env:
            env_vars = [[k, v] for k, v in env.items()]
            self._load_env_vars(env_vars)

    def list_tasks(self):
        """List all available tasks."""
        if all(t.is_builtin for t in self._tasks.values()):
            self.print_error("No tasks have been registered. plz expects at least one `@plz.task` in your plzfile.py")
            return

        max_command_length = max(len(t.name) + 5 for t in self._tasks.values())
        max_command_length = min(max_command_length, 25)

        max_desc_length = max(len(t.desc) * 2 for t in self._tasks.values() if t.desc is not None)
        max_desc_length = min(max_desc_length, 50)

        table = Table(show_header=False, box=None, show_edge=False)
        table.add_column("Tasks", style="orange1", no_wrap=True, width=max_command_length + 2)
        table.add_column("Description", style="white", no_wrap=True, width=max_desc_length + 2)

        for t in self._tasks.values():
            desc = t.desc or ""
            name = t.name
            if t.is_default:
                name = f"[bold]{name}[/bold]"
                desc = f"[bold]{desc}\t(default)[/bold]"
            table.add_row(f"{name}", desc)

        panel_width = max_command_length + max_desc_length + 4

        # Ensure the panel width does not exceed terminal width
        terminal_width = console.size.width

        final_width = min(panel_width, terminal_width)

        panel = Panel(
            table,
            title="Tasks",
            title_align="left",
            border_style="dark_orange3",
            padding=(0, 1),
            box=box.ROUNDED,
            width=final_width,
        )
        console.print(panel)

    def _get_dotenv_path(self) -> Path:
        return Path(os.getcwd()) / ".env"

    def _load_env_dotenv(self) -> list[list[str]]:
        env_path = self._get_dotenv_path()

        # Load the .env file if it exists
        if env_path.exists():
            # capture the env variables
            load_dotenv(env_path)
            # Load the environment variables into a variable
            env_vars_dict = dotenv_values(env_path)
            env_vars_list = [[k, v] for k, v in env_vars_dict.items() if v is not None]
            return env_vars_list

        return []

    def _try_execute_utility_commands(self, command: Command) -> bool:
        if command.has_task_specified():
            return False

        if command.list:
            self.list_tasks()
            return True

        if command.list_env:
            self._print_env(cmd=command)
            return True

        if command.list_env_all:
            self._print_env(cmd=command, all=True)
            return True

        return False

    def _get_default_task(self) -> Task | None:
        default_tasks = [t for t in self._tasks.values() if t.is_default]

        if len(default_tasks) > 1:
            self.fail("More than one default task found: " + ", ".join(t.name for t in default_tasks))

        if len(default_tasks) == 0:
            return None
        return default_tasks[0]

    def _try_execute_default_task(self, command: Command) -> bool:
        if command.has_task_specified():
            return False

        if not command.is_default():
            return False

        default_task = self._get_default_task()
        if default_task is None:
            self.list_tasks()
            return True
        else:
            default_task()
            return True

    def _try_execute_task(self, command: Command) -> bool:
        if not command.has_task_specified():
            return False

        assert command.task is not None

        if command.list_env:
            self._print_env(cmd=command)
            return True

        kwargs = command.task_kwargs or {}
        command.task(**kwargs)
        return True

    def _load_env_cli(self, command: Command):
        self._load_env_vars(command.env)

    def _load_env_vars(self, env_vars: list[list[str]]):
        for k, v in env_vars:
            os.environ[k] = v

    def _load_plzfile(self):
        plzfile_path = os.path.join(os.getcwd(), "plzfile.py")
        spec = importlib.util.spec_from_file_location("plzfile", plzfile_path)
        plzfile = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(plzfile)  # type: ignore

    def _main_execute(self, command: Command):
        if not self._user_configured:
            self.configure()

        self._load_env_cli(command)

        if self._try_execute_utility_commands(command):
            return

        if self._try_execute_default_task(command):
            return

        if self._try_execute_task(command):
            return

        self.fail("Execution failed for unknown reason.")

    def task(
        self,
        name: str | None = None,
        desc: str | None = None,
        default: bool = False,
        requires: Callable | list[Callable | CallableWithArgs] | None = None,
        env: dict[str, Any] | None = None,
    ) -> Callable:
        """Defines a plz task that can be executed from the command line.

        Functions decorated with `plz.task` will be registered as tasks that can be executed from the command line.
        They can also accept arguments which will get documented in the task's help message.
        Also, you can declare *args to accept any number of positional arguments.

        Args:
            name (str, optional): _description_. Defaults to the method name (with underscores replaced with hyphens).
            desc (str, optional): _description_. Defaults to the method's docstring.
            default (bool, optional): _description_. Defaults to False. If True, this task will be executed if no task
                explicilty specified.
            requires (Callable | list[Callable  |  CallableWithArgs] | None, optional): _description_. Defaults to None.
                Defines the required tasks that will be executed before this task. Can be a single task (functions),
                or a list of tasks (optionally with arguments). See examples below, or refrence the docs for more
                information.
            env (dict[str, Any] | None, optional): _description_. Defaults to None. Defines task-level environment
                variables that will be set before the task is executed.
        """

        def decorator(func) -> Callable:
            t_name: str = name or func.__name__

            t_desc = desc
            if desc is None:
                t_desc = inspect.cleandoc(func.__doc__) if func.__doc__ else ""

            # Nomralize requires to list of tuples
            _required = requires
            required_funcs: list[CallableWithArgs]
            if _required is None:
                required_funcs = []
            elif isinstance(_required, list):
                # Normalize callable with args to tuple as well
                required_funcs = [r if isinstance(r, tuple) else (r, ()) for r in _required]
            else:
                required_funcs = [(_required, ())]
            required_tasks = [(self._tasks[r.__name__], args) for r, args in required_funcs]

            self._tasks[func.__name__] = Task(
                func=func, name=t_name, desc=t_desc, is_default=default, requires=required_tasks, task_env_vars=env
            )

            return func

        return decorator

    @staticmethod
    def print_error(msg: str, silent: bool = False, exit: bool = False):
        PlzApp.print(msg, "red", silent=silent)
        if exit:
            sys.exit(1)

    @staticmethod
    def print_warning(msg: str, silent: bool = False):
        PlzApp.print(msg, "yellow", silent=silent)

    @staticmethod
    def print_weak(msg: str, silent: bool = False):
        PlzApp.print(msg, "bright_black", silent=silent)

    @staticmethod
    def print(msg: str, color: str | None = None, silent: bool = False):
        if silent:
            return

        if color:
            msg = f"[{color}]{msg}[/]"
        console.print(msg)

    @staticmethod
    def fail(msg: str | None = None):
        if msg is not None:
            PlzApp.print_error(msg)
        sys.exit(1)

    def _print_env(self, cmd: Command, all: bool = False):
        """
        prints the environment variable
        """
        ConsoleUtils.print_box(title=".env", rows=self._dotenv, sort=True)
        ConsoleUtils.print_box(title="in-line", rows=cmd.env, sort=True)
        if all:
            env_vars = os.environ
            rows = [[key, value] for key, value in env_vars.items()]
            rows = [row for row in rows if row not in self._dotenv]
            ConsoleUtils.print_box(title="All (rest)", rows=rows, sort=True)

    def run(
        self,
        command: str,
        env: dict[str, str] | None = None,
        dry_run: bool = False,
        echo: bool = True,
        raise_error: bool = True,
    ) -> int:
        """
        Executes a shell command with optional environment variables, timeout, and dry run mode, and more.
        Args:
            command (str): The shell command to execute.
            env (dict[str, str], optional): A dictionary of environment variables to set for the command.
                Defaults to None.
            dry_run (bool): If True, the command will not be executed, and a dry run message will be printed.
                Defaults to False.
            echo (bool): If True, the command will be printed before execution (with environment variables replaced).
                Defaults to True.
            raise_error (bool): If True, a CalledProcessError will be raised if the command returns a non-zero
                exit status. Defaults to True.
        Returns:
            The standard output of the command if it was successful or the standard error if it failed.
        Raises:
            subprocess.CalledProcessError: If the command returns a non-zero exit status.
        """
        # Merge provided env with the current environment variables
        env = {**os.environ, **(env or {})}

        # replace env variables in the command
        def replace_env_var(match):
            var_name = match.group(1)
            return env.get(var_name, f"${{{var_name}}}")

        command_w_vars = re.sub(r"\$(\w+)", replace_env_var, command)

        if echo and not dry_run:
            self.print_weak(f"Executing: `{command_w_vars}`")

        if dry_run:
            self.print_warning(f"Dry run: `{command_w_vars}`")
            return 0

        ec = os.system(command_w_vars)

        if raise_error and ec != 0:
            raise subprocess.CalledProcessError(ec, command)

        return ec

    def ask(self, question: str | None = None, silent: bool = False) -> bool:
        """
        Asks the user a yes/no question and returns a boolean based on the response.

        Args:
            question (str): The question to ask the user. Defaults to "Are you sure [y] / [n]".
            silent (bool): If True, response will not be declared. Defaults to False.
        Returns:
            bool: True if the user responds with 'y' or 'Y', False otherwise.
        """
        if question is None:
            question = "Are you sure you want to proceed? (y/n) [n]:"
        while True:
            response = input(f"{question} ").strip().lower()
            if response in ["y", "n", "Y", "N", ""]:
                if response.lower() == "y":
                    self.print_weak("Accepted", silent=silent)
                    return True
                else:
                    self.print_weak("Declined", silent=silent)
                    return False
            self.print_warning("Please respond with 'y' or 'n'. Enter will default to 'n'.")


plz = PlzApp()



from pyplz.command import Parser


class TestParser:
    def test_parser_list(self):
        parser = Parser()

        cmd = parser.parse_args(["-l"])

        assert cmd.list
        assert not cmd.has_task_specified()
        assert not cmd.is_default()

    def test_parser_list_full(self):
        parser = Parser()

        cmd = parser.parse_args(["--list"])

        assert cmd.list
        assert not cmd.has_task_specified()
        assert not cmd.is_default()

    def test_parser_help(self):
        parser = Parser()

        cmd = parser.parse_args(["-h"])

        assert cmd.help
        assert not cmd.has_task_specified()
        assert not cmd.is_default()

    def test_parser_help_full(self):
        parser = Parser()

        cmd = parser.parse_args(["--help"])

        assert cmd.help
        assert not cmd.has_task_specified()
        assert not cmd.is_default()

    def test_parser_empty(self):
        parser = Parser()

        cmd = parser.parse_args([])

        assert not cmd.has_task_specified()
        assert cmd.is_default()

    def test_parser_task(self):
        parser = Parser()

        cmd = parser.parse_args(["some-task"])

        assert cmd.has_task_specified()
        assert cmd.args == []
        assert not cmd.is_default()

    def test_parser_task_with_help(self):
        parser = Parser()

        cmd = parser.parse_args(["some-task", "-h"])

        assert cmd.has_task_specified()
        assert cmd.args == []
        assert not cmd.is_default()
        assert cmd.help

    def test_parser_task_with_args(self):
        parser = Parser()

        cmd = parser.parse_args(["some-task", "arg1", "arg2"])

        assert cmd.has_task_specified()
        assert cmd.args == ["arg1", "arg2"]
        assert not cmd.is_default()

    def test_parser_task_with_args_and_env(self):
        parser = Parser()

        cmd = parser.parse_args(["-e" "a=1", "some-task", "arg1", "arg2"])

        assert cmd.has_task_specified()
        assert cmd.args == ["arg1", "arg2"]
        assert cmd.env == [["a", "1"]]
        assert not cmd.is_default()

    def test_parser_task_with_args_and_env_mixed(self):
        parser = Parser()

        cmd = parser.parse_args(["some-task", "arg1", "-e", "a=1", "arg2"])

        assert cmd.has_task_specified()
        assert cmd.args == ["arg1", "arg2"]
        assert cmd.env == [["a", "1"]]
        assert not cmd.is_default()



from __future__ import annotations

from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

console = Console()


class ConsoleUtils:
    @staticmethod
    def print_box(title: str, rows: list[list[str]], sort: bool = False) -> None:
        if rows is None or len(rows) == 0:
            return

        # assert only 2 columns (only use case for now)
        if any(len(r) != 2 for r in rows):
            raise ValueError("Only supporting 2 columns")

        if sort:
            rows = sorted(rows, key=lambda x: x[0])

        # get the max length between all first items
        max_first_length = max(len(r[0]) for r in rows)
        max_first_length = max(max_first_length, 15)

        max_seoncd_length = max(len(r[1]) for r in rows)
        max_seoncd_length = max(max_seoncd_length, 15)

        table = Table(show_header=False, box=None, show_edge=False)
        table.add_column("1", no_wrap=True, width=max_first_length + 2)
        table.add_column("2", style="white", no_wrap=True, width=max_seoncd_length + 2)

        for r in rows:
            first = f"[bold]{r[0]}[/]" or ""
            second = r[1] or ""
            table.add_row(first, second)

        panel_width = max_first_length + max_seoncd_length + 4

        # Ensure the panel width does not exceed terminal width
        terminal_width = console.size.width

        final_width = min(panel_width, terminal_width)

        panel = Panel(
            table,
            title=title,
            title_align="left",
            border_style="cyan",
            padding=(0, 1),
            box=box.ROUNDED,
            width=final_width,
        )
        console.print(panel)



from __future__ import annotations

import os
import sys
import tempfile
from pathlib import Path
from unittest.mock import Mock

from pyplz import plz
from pyplz.command import Command
from pyplz.main import main
from tests.conftest import TestUtils


class TestEnv:
    @TestUtils.patch_method(plz._get_dotenv_path)
    def test_dotenv_loaded(self, get_dotenv_path_mock):
        dotenv_content = "KEY1=value1\nKEY2=value2\nKEY3=value3"
        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(dotenv_content.encode())
            tmp.seek(0)
            get_dotenv_path_mock.return_value = Path(tmp.name)
            plz._load_env_dotenv()
            assert os.getenv("KEY1") == "value1"
            assert os.getenv("KEY2") == "value2"
            assert os.getenv("KEY3") == "value3"

    @TestUtils.patch_method(plz._get_dotenv_path)
    def test_dotenv_loaded_e2e(self, get_dotenv_path_mock):
        dotenv_content = "KEY1=value1\nKEY2=value2\nKEY3=value3"
        impl_mock = Mock()

        @plz.task()
        def sample_task():
            assert os.getenv("KEY1") == "value1"
            assert os.getenv("KEY2") == "value2"
            assert os.getenv("KEY3") == "value3"
            impl_mock()

        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(dotenv_content.encode())
            tmp.seek(0)
            get_dotenv_path_mock.return_value = Path(tmp.name)

            sys.argv = ["plz", "sample_task"]

            main()

        impl_mock.assert_called_once()

    @TestUtils.patch_method(plz._load_env_dotenv)
    def test_environment_variables_loaded(self, load_environment_variables_mock):
        cmd = Command()
        plz._main_execute(cmd)
        load_environment_variables_mock.assert_called_once()

    @TestUtils.patch_method(plz._load_env_cli)
    def test_inline_env_vars_loading_called(self, process_env_vars_mock):
        cmd = Command()
        plz._main_execute(cmd)
        process_env_vars_mock.assert_called_once()

    def test_inline_env_vars_loading_loaded(self):
        cmd = Command(None, _env=["KEY1=value1", "KEY2=value2", "KEY3=value3"])
        plz._main_execute(cmd)
        assert os.getenv("KEY1") == "value1"
        assert os.getenv("KEY2") == "value2"
        assert os.getenv("KEY3") == "value3"

    def test_inline_env_vars_loading_loaded_e2e(self):
        impl_mock = Mock()

        @plz.task()
        def sample_task():
            assert os.getenv("KEY1") == "value1"
            assert os.getenv("KEY2") == "value2"
            assert os.getenv("KEY3") == "value3"
            impl_mock()

        sys.argv = ["plz", "-e", "KEY1=value1", "-e", "KEY2=value2", "-e", "KEY3=value3", "sample_task"]

        main()

        impl_mock.assert_called_once()

    def test_task_env_vars_loaded(self):
        impl_mock = Mock()

        @plz.task(env={"a": "1"})
        def sample_task():
            assert os.getenv("a") == "1"
            impl_mock()

        plz._main_execute(Command(plz._tasks["sample_task"]))
        impl_mock.assert_called_once()

    def test_task_env_vars_loaded_e2e(self):
        impl_mock = Mock()

        @plz.task(env={"a": "1"})
        def sample_task():
            assert os.getenv("a") == "1"
            impl_mock()

        sys.argv = ["plz", "sample_task"]
        main()

        impl_mock.assert_called_once()

    def test_task_env_configured(self):
        impl_mock = Mock()

        plz.configure(env={"a": "1", "b": "2"})

        @plz.task()
        def sample_task():
            assert os.getenv("a") == "1"
            assert os.getenv("b") == "2"
            impl_mock()

        sys.argv = ["plz", "sample_task"]
        main()

        impl_mock.assert_called_once()



from __future__ import annotations

from unittest.mock import Mock

from pyplz import plz
from pyplz.command import Command
from pyplz.task import Task


class TestPlzHelp:
    def test_help_global(self):
        cmd = Command(help=True)
        print_help_mock = Mock()
        plz._print_help = print_help_mock

        plz._main_execute(cmd)

        print_help_mock.assert_called_once()

    def test_help_task(self):
        print_help_mock = Mock()
        sample_task_mock = Mock()
        Task.print_help = print_help_mock

        @plz.task()
        def sample_task():
            sample_task_mock()

        cmd = Command(task="sample_task", help=True)

        plz._main_execute(cmd)

        print_help_mock.assert_called_once()
        sample_task_mock.assert_not_called()



from __future__ import annotations

import argparse
from dataclasses import dataclass
from typing import List

from pyplz.task import Task


@dataclass
class Command:
    task: Task | None = None
    task_kwargs: dict[str, str] | None = None
    list: bool = False
    help: bool = False
    list_env: bool = False
    list_env_all: bool = False
    _env: List[str] | None = None
    _args: List[str] | None = None

    def has_task_specified(self) -> bool:
        return self.task is not None

    def is_default(self) -> bool:
        has_any_utility_flag = self.list or self.help or self.list_env or self.list_env_all
        return not self.has_task_specified() and not has_any_utility_flag

    @property
    def env(self) -> List[List[str]]:
        return [env.split("=") for env in self._env] if self._env else []

    @property
    def args(self) -> List[str]:
        return self._args or []


class Parser:
    def __init__(self):
        self.parser = argparse.ArgumentParser(description="plz - A python-first task runner.", add_help=False)
        self.parser.add_argument("task", nargs="*", help="The command to run")
        # self.parser.add_argument("-l", "--list", action="store_true", help="List all available tasks")
        # self.parser.add_argument(
        #     "-h",
        #     "--help",
        #     action="store_true",
        #     help="Show help for a specific task (or for plz if no task is provided)",
        # )
        # self.parser.add_argument(
        #     "--list-env",
        #     action="store_true",
        #     help="List dedicated environment variables for the task (or for plz if no task is provided)",
        # )
        # self.parser.add_argument("--list-env-all", action="store_true", help="List all environment variables")
        # self.parser.add_argument(
        #     "-e",
        #     "--env",
        #     action="append",
        #     type=str,
        #     help="Set an environment variable (can be used multiple times). Example: -e KEY=VALUE",
        # )
        # self.parser.add_argument("args", nargs=argparse.REMAINDER, help="Additional arguments for the task")

        self.task_parser = argparse.ArgumentParser(add_help=False)

    def _split_env_vars_from_args(self, args: List[str]) -> tuple[List[str], List[str]]:
        env_vars = []
        arguments = []
        i = 0
        while i < len(args):
            if (args[i] == "-e" or args[i] == "--env") and i + 1 < len(args):
                env_vars.append(args[i + 1])
                i += 2
            else:
                arguments.append(args[i])
                i += 1
        return env_vars, arguments

    def parse_args(self, argv: list[str]) -> Command:
        args = self.parser.parse_args(argv)

        generic_task_flags = ["-h", "--help", "--list-env"]
        clean_args = [arg for arg in args.args if arg not in generic_task_flags]

        task_env_vars, arguments = self._split_env_vars_from_args(clean_args)

        # join both env vars list
        args.env = args.env or []
        args.env.extend(task_env_vars)

        return Command(
            task=args.task,
            list=args.list,
            help=args.help,
            list_env=args.list_env,
            list_env_all=args.list_env_all,
            _env=args.env,
            _args=arguments,
        )



import argparse

from pyplz.command import Command
from pyplz.exceptions import TaskNotFoundError
from pyplz.plz_app import PlzApp
from pyplz.task import Task
from pyplz.task_parser import TaskParser


class PlzParser:
    def __init__(self, plz_app: PlzApp, task_parser: TaskParser):
        self.plz_app = plz_app
        self.task_parser = task_parser
        self.parser = argparse.ArgumentParser(description="plz - A python-first task runner.", add_help=True)
        self.parser.add_argument("-l", "--list", action="store_true", help="List all available tasks")
        self.parser.add_argument("-e", "--env", action="append", help="Inline environment variable (KEY=VALUE)")
        self.parser.add_argument("task", nargs=argparse.REMAINDER, help="The task to run")

    def parse(self, args: list[str]) -> Command:
        parsed_args = self.parser.parse_args(args)

        # Parse inline environment variables if provided.
        if parsed_args.task:
            cmd = self.parse_task(parsed_args.task)
            cmd._env = parsed_args.env
        else:
            cmd = Command(list=parsed_args.list, _env=parsed_args.env)

        return cmd

    def list_tasks(self):
        # Implement the logic to list all available tasks
        print("Listing all available tasks...")

    def parse_task(self, task_with_args: list[str]) -> Command:
        task_name = task_with_args[0]
        if task_name not in self.plz_app._tasks:
            raise TaskNotFoundError(task_name=task_name)
        if len(task_with_args) > 1:
            task_args = task_with_args[1:]
        else:
            task_args = []

        task: Task = self.plz_app._tasks[task_name]
        # Implement the logic to run the task with the provided arguments

        cmd = self.task_parser.parse(task=task, args=task_args)

        return cmd



from __future__ import annotations

import subprocess

import pytest

from pyplz import plz


class TestPlzRun:
    def test_run_fails_error_raised(self):
        with pytest.raises(subprocess.CalledProcessError):
            plz.run("ls /non/existent/path", raise_error=True)

    def test_run_fails_error_not_raised(self):
        ec = plz.run("ls /non/existent/path", raise_error=False)
        assert ec != 0

    def test_run_echo_true(self, capfd):
        cmd = "echo hello"

        ec = plz.run(cmd, echo=True)

        lines = capfd.readouterr().out.splitlines()
        assert len(lines) == 2
        assert lines[0] == f"Executing: `{cmd}`"
        assert lines[1] == "hello"
        assert ec == 0

    def test_run_echo_false(self, capfd):
        cmd = "echo hello"

        ec = plz.run(cmd, echo=False)

        lines = capfd.readouterr().out.splitlines()
        assert len(lines) == 1
        assert lines[0] == "hello"
        assert ec == 0
